import asyncio
import json
import logging
import time
import websockets

class Node_WS:
    def __init__(self, debug, nodeid, nodetype, maxlevel, ws):
        self.debug = debug
        self.nodeid = nodeid
        self.nodetype = nodetype
        self.maxlevel = maxlevel
        self.ws = ws
        self.status = {}
        ## logging
        FORMAT='[%(asctime)13s :: %(name)18s :: %(levelname)7s] %(message)s'
        DATEFMT='%Y%m%d %H:%M:%S'
        logging.basicConfig(level=logging.INFO, format=FORMAT,
                datefmt=DATEFMT)
        self.logger = logging.getLogger(f"Node_WS:{self.nodeid.split('-')[0]}")
        if self.debug:
            self.logger.setLevel(logging.DEBUG)

    def get_info(self):
        ret = {}
        ret["nodeid"] = self.nodeid
        ret["nodetype"] = self.nodetype
        ret["maxlevel"] = self.maxlevel
        return ret

    def get_status(self):
        ret = self.status
        return ret

class Core_WS:
    def __init__(self, debug, host, port):
        self.debug = debug
        self.host = host
        self.port = port
        self.nodes = {}
        ## logging
        FORMAT='[%(asctime)13s :: %(name)18s :: %(levelname)7s] %(message)s'
        DATEFMT='%Y%m%d %H:%M:%S'
        logging.basicConfig(level=logging.INFO, format=FORMAT,
                datefmt=DATEFMT)
        self.logger = logging.getLogger("Core_WS")
        if self.debug:
            self.logger.setLevel(logging.DEBUG)
        self.ws_server = websockets.serve(self.handler, self.host, self.port)

    def get_nodes(self):
        ret = {}
        for n in self.nodes:
            ret[n] = self.nodes[n].get_info()
        return ret

    def get_node(self, nodeid):
        ret = None
        if nodeid in self.nodes:
            ret = self.nodes[nodeid].get_info()
        return ret


    async def node_register(self, nodeid, nodetype, maxlevel, websocket):
        self.nodes[nodeid] = Node_WS(self.debug, nodeid, nodetype, maxlevel, websocket)
        self.logger.info(f"node registered: {nodeid}")

    async def node_unregister(self, nodeid):
        if nodeid and nodeid in self.nodes:
            self.nodes.pop(nodeid)
            self.logger.info(f"node unregistered: {nodeid.split('-')[0]}")

    async def node_send_command(self, nodeid, command, data):
        if nodeid in self.nodes:
            m = {"command": command, "data": data}
            message = json.dumps(m)
            self.logger.debug(f"> {nodeid} {command}")
            if self.nodes[nodeid].ws.open:
                await asyncio.wait([self.nodes[nodeid].ws.send(message)])
            else:
                self.logger.warning(f"send_command({nodeid}): ws not open")
                self.node_unregister(nodeid)
        else:
            self.logger.warning(f"send_command({nodeid}) unknown")

    async def handler(self, websocket, path):
        nodeid = None
        try:
            async for message in websocket:
                m = json.loads(message)
                # self.logger.debug(f"< {m}")

                nodeid = m.get("nodeid")
                cmd = m.get("command")
                if nodeid and cmd:
                    self.logger.debug(f"< {nodeid} {cmd}")
                    if cmd == "register":
                        if not nodeid in self.nodes:
                            nodetype = m.get("nodetype")
                            maxlevel = m.get("maxlevel")
                            await self.node_register(nodeid, nodetype, maxlevel,
                                    websocket)
                            ret = {"command": "registered"}
                            await websocket.send(json.dumps(ret))
                    elif cmd == "status":
                        status = m.get("status")
                        if status:
                            self.nodes[nodeid].status = status
                            self.nodes[nodeid].status["ts"] = time.time()
                    else:
                        self.logger.warning(f"unknown command: {m}")

        except websockets.ConnectionClosed:
            pass

        finally:
            await self.node_unregister(nodeid)
